/* Copyright Euresys 2014 */

#ifndef COAXLINKGRABBER_HEADER_FILE
#define COAXLINKGRABBER_HEADER_FILE

#include <string>
#include <vector>
#include <algorithm>
#include <EuresysGenTL.h>
#include <EuresysLocks.h>

// namespace gc = GenICam::Client;

struct InterfacePort {
};

struct DevicePort {
};

struct StreamPort {
};

struct RemotePort {
};

class CoaxlinkGrabber {
    public:
        CoaxlinkGrabber(Euresys::GenTL &gentl, int interfaceIndex = 0, int deviceIndex = 0, int dataStreamIndex = 0)
        : gentl(gentl)
        , tlh(GENTL_INVALID_HANDLE)
        , ifh(GENTL_INVALID_HANDLE)
        , devh(GENTL_INVALID_HANDLE)
        , camh(GENTL_INVALID_HANDLE)
        , dsh(GENTL_INVALID_HANDLE)
        , eh(GENTL_INVALID_HANDLE)
        , buffers(0)
        , width(0)
        , height(0)
        , payloadSize(0)
        , usingUserBuffers(false)
        , started(false)
        , cameraStarted(false)
        {
            tlh = gentl.tlOpen();
            gentl.tlUpdateInterfaceList(tlh);
            ifh = gentl.tlOpenInterface(tlh, gentl.tlGetInterfaceID(tlh, interfaceIndex));
            gentl.ifUpdateDeviceList(ifh);
            devh = gentl.ifOpenDevice(ifh, gentl.ifGetDeviceID(ifh, deviceIndex));
            camh = gentl.devGetPort(devh);
            dsh = gentl.devOpenDataStream(devh, gentl.devGetDataStreamID(devh, dataStreamIndex));
            eh = gentl.gcRegisterEvent(dsh, Euresys::gc::EVENT_NEW_BUFFER);
            ReallocBuffers(0);
        }

        virtual ~CoaxlinkGrabber() {
            try {
                gentl.genapiExecuteCommand(camh, "AcquisitionStop");
            } catch (...) {
            }
            try {
                gentl.devClose(devh);
            } catch (...) {
            }
            try {
                gentl.ifClose(ifh);
            } catch (...) {
            }
            try {
                gentl.tlClose(tlh);
            } catch (...) {
            }
        }

        void memento(const std::string &text) {
            gentl.memento(text);
        }

    private:
        bool UpdateConfig(size_t buffer_count, size_t buffer_size) {
            int64_t camera_width = gentl.genapiGetInteger(camh, "Width");
            int64_t camera_height = gentl.genapiGetInteger(camh, "Height");
            std::string camera_pixelFormat = gentl.genapiGetString(camh, "PixelFormat");
            int64_t camera_pitch = camera_width * evalPixelSize(camera_pixelFormat);
            if (camera_pitch <= 0) {
                throw Euresys::client_error(std::string("Invalid pitch"));
            }

            if (buffer_size == 0) {
                int64_t camera_payloadSize = -1;
                if (camera_payloadSize == -1) {
                    camera_payloadSize = gentl.genapiGetIntegerWithDefault(camh, "PayloadSize", -1);
                }
                if (camera_payloadSize == -1) {
                    camera_payloadSize = gentl.genapiGetIntegerWithDefault(camh, "PayLoadSize", -1);
                }
                if (camera_payloadSize == -1) {
                    camera_payloadSize = evalPayloadSize(camera_width, camera_height, camera_pixelFormat);
                }
                buffer_size = static_cast<size_t>(camera_payloadSize);
            }

            bool newBufferCount = (buffers.size() != buffer_count);
            bool newPayloadSize = (static_cast<size_t>(buffer_size) != payloadSize);
            bool changed = usingUserBuffers || newBufferCount || newPayloadSize;

            width = camera_width;
            height = buffer_size / camera_pitch;
            pixelFormat = camera_pixelFormat;
            payloadSize = buffer_size;

            if (changed) {
                // revoke current buffers
                gentl.dsFlushQueue(dsh, Euresys::gc::ACQ_QUEUE_ALL_DISCARD);
                gentl.dsRevokeBuffers(dsh, buffers);
                buffers.resize(buffer_count, GENTL_INVALID_HANDLE);
                usingUserBuffers = false;
            }
            return changed;
        }

    public:
        bool ReallocBuffers(size_t buffer_count, size_t buffer_size = 0) {
            Euresys::AutoLock lock(mutex);
            if (!UpdateConfig(buffer_count, buffer_size)) {
                return false;
            }
            usingUserBuffers = false;
            gentl.dsAllocAndAnnounceBuffers(dsh, payloadSize, buffers);
            gentl.dsQueueBuffers(dsh, buffers);
            return true;
        }

        void AnnounceBuffer(void *ptr, size_t size, void *userPointer = 0) {
            Euresys::AutoLock lock(mutex);
            usingUserBuffers = true;
            buffers.push_back(gentl.dsAnnounceBuffer(dsh, ptr, size, userPointer));
            gentl.dsQueueBuffer(dsh, buffers.back());
        }

        void Start(uint64_t frame_count = GENTL_INFINITE) {
            Euresys::AutoLock lock(mutex);
            if (!cameraStarted) {
                gentl.genapiExecuteCommand(camh, "AcquisitionStart");
                cameraStarted = true;
            }
            if (started) {
                gentl.dsStopAcquisition(dsh, Euresys::gc::ACQ_STOP_FLAGS_KILL);
                started = false;
            }
            gentl.dsStartAcquisition(dsh, Euresys::gc::ACQ_START_FLAGS_DEFAULT, frame_count);
            started = true;
        }

        void Stop() {
            Euresys::AutoLock lock(mutex);
            gentl.dsStopAcquisition(dsh, Euresys::gc::ACQ_STOP_FLAGS_DEFAULT);
            started = false;
            gentl.genapiExecuteCommand(camh, "AcquisitionStop");
            cameraStarted = false;
        }

        class Buffer {
            public:
                Buffer(Euresys::gc::DS_HANDLE dsh, Euresys::gc::BUFFER_HANDLE bh, void *userPointer)
                : dsh(dsh)
                , bh(bh)
                , userPointer(userPointer)
                {}
                void Push(CoaxlinkGrabber &grabber) {
                    grabber.gentl.dsQueueBuffer(dsh, bh);
                }
                std::string getInfoString(CoaxlinkGrabber &grabber, Euresys::gc::BUFFER_INFO_CMD cmd) {
                    return grabber.gentl.dsGetBufferInfoString(dsh, bh, cmd);
                }
                template<typename T> T getInfo(CoaxlinkGrabber &grabber, Euresys::gc::BUFFER_INFO_CMD cmd) {
                    return grabber.gentl.dsGetBufferInfo<T>(dsh, bh, cmd);
                }
                void *GetUserPointer() {
                    return userPointer;
                }
            private:
                Euresys::gc::DS_HANDLE dsh;
                Euresys::gc::BUFFER_HANDLE bh;
                void *userPointer;
        };

        class ScopedBuffer {
            public:
                ScopedBuffer(CoaxlinkGrabber &grabber, uint64_t timeout = GENTL_INFINITE)
                : grabber(grabber)
                , buffer(grabber.Pop(timeout))
                {}
                ~ScopedBuffer() {
                    try {
                        buffer.Push(grabber);
                    } catch (...) {
                    }
                }
                std::string getInfoString(Euresys::gc::BUFFER_INFO_CMD cmd) {
                    return buffer.getInfoString(grabber, cmd);
                }
                template<typename T> T getInfo(Euresys::gc::BUFFER_INFO_CMD cmd) {
                    return buffer.getInfo<T>(grabber, cmd);
                }
                void *GetUserPointer() {
                    return buffer.GetUserPointer();
                }
            protected:
                static void *operator new(size_t); // no new ScopedBuffer
                static void *operator new[](size_t);  // no new ScopedBuffer[]
            private:
                ScopedBuffer(const ScopedBuffer &); // no copy
                ScopedBuffer &operator=(const ScopedBuffer &); // no assignment

                CoaxlinkGrabber &grabber;
                Buffer buffer;
        };

        Buffer Pop(uint64_t timeout = GENTL_INFINITE) {
            Euresys::gc::EVENT_NEW_BUFFER_DATA data;
            gentl.eventGetData(eh, &data, sizeof(data), timeout);
            return Buffer(dsh, data.BufferHandle, data.pUserPointer);
        }

        int64_t GetWidth() const {
            return width;
        }

        int64_t GetHeight() const {
            return height;
        }

        size_t GetPayloadSize() const {
            return payloadSize;
        }

        std::string GetPixelFormat() const {
            return pixelFormat;
        }

        template<typename P> int64_t getInteger(const std::string &feature) {
            return gentl.genapiGetInteger(getPort((P *)0), feature);
        }
        template<typename P> double getFloat(const std::string &feature) {
            return gentl.genapiGetFloat(getPort((P *)0), feature);
        }
        template<typename P> std::string getString(const std::string &feature) {
            return gentl.genapiGetString(getPort((P *)0), feature);
        }
        template<typename P> std::vector<std::string> getStringList(const std::string &what) {
            return gentl.genapiGetStringList(getPort((P *)0), what);
        }
        template<typename P> void setInteger(const std::string &feature, int64_t value) {
            gentl.genapiSetInteger(getPort((P *)0), feature, value);
        }
        template<typename P> void setFloat(const std::string &feature, double value) {
            gentl.genapiSetFloat(getPort((P *)0), feature, value);
        }
        template<typename P> void setString(const std::string &feature, const std::string &value) {
            gentl.genapiSetString(getPort((P *)0), feature, value);
        }
        template<typename P> void execute(const std::string &command) {
            gentl.genapiExecuteCommand(getPort((P *)0), command);
        }

        void runScript(const std::string &script) {
            gentl.genapiRunScript(dsh, script);
        }

    private:
        Euresys::gc::PORT_HANDLE getPort(InterfacePort *) {
            return ifh;
        }
        Euresys::gc::PORT_HANDLE getPort(DevicePort *) {
            return devh;
        }
        Euresys::gc::PORT_HANDLE getPort(StreamPort *) {
            return dsh;
        }
        Euresys::gc::PORT_HANDLE getPort(RemotePort *) {
            return camh;
        }

        static int evalPixelSize(const std::string &pixelFormat) {
            std::string format(pixelFormat);
            std::transform(format.begin(), format.end(), format.begin(), ::tolower);
            if (format == "mono8")      return 1;
            if (format == "mono10")     return 2;
            if (format == "mono12")     return 2;
            if (format == "mono14")     return 2;
            if (format == "mono16")     return 2;
            if (format == "bayergb8")   return 1;
            if (format == "bayerbg8")   return 1;
            if (format == "bayergr8")   return 1;
            if (format == "bayerrg8")   return 1;
            if (format == "rgb8")       return 3;
            if (format == "rgba8")      return 4;
            if (format == "rgb8a32")    return 4;
            if (format == "bgr8")       return 3;
            if (format == "bgra8")      return 4;
            if (format == "bgr8a32")    return 4;
            throw Euresys::client_error(std::string("Unknown PixelFormat ") + pixelFormat);
        }

        static size_t evalPayloadSize(int64_t width, int64_t height, const std::string &pixelFormat) {
            return static_cast<size_t>(width * height * evalPixelSize(pixelFormat));
        }

        Euresys::ConcurrencyLock mutex;
        Euresys::GenTL &gentl;
        Euresys::gc::TL_HANDLE tlh;
        Euresys::gc::IF_HANDLE ifh;
        Euresys::gc::DEV_HANDLE devh;
        Euresys::gc::PORT_HANDLE camh;
        Euresys::gc::DS_HANDLE dsh;
        Euresys::gc::EVENT_HANDLE eh;
        std::vector<Euresys::gc::BUFFER_HANDLE> buffers;
        int64_t width;
        int64_t height;
        size_t payloadSize;
        std::string pixelFormat;
        bool usingUserBuffers;
        bool started;
        bool cameraStarted;
};

#endif
